#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/param.h>
#include <assert.h>

#include "config_unix.h"
#include "config_win32.h"
#include "net_udp.h"
#include "debug.h"

#include "config.h"
#include "util.h"

struct shared {
        u_int32 packets_sent;
        u_int32 reports_recv;
        u_int32 rtt_cur;
        double  rtt;
        double  sqrt_rtt;
        double  b_rep;
        double  b_act;
        double  b_exp;
        double  min_bitrate;
        double  max_bitrate;
        double  max_increase;
        double  rate_delta;
        int64   adj_ts;
        int64   rep_ts;
        u_int32 round;
        int     mode;
        int     force_n_mode;
};


int shm_init(struct shared *info, double minb, double maxb) {
        info->packets_sent    = 0;
        info->reports_recv    = 0;
        info->rtt_cur         = 0;
        info->rtt             = 0.0;
        info->sqrt_rtt        = 0.0;
        info->b_rep           = 0.0;
        info->b_exp           = PKT_SIZE * 2;
        info->b_act           = info->b_exp;
        info->min_bitrate     = minb;
        info->max_bitrate     = maxb;
        info->max_increase    = info->b_exp;
        info->rate_delta      = 0.0;
        info->adj_ts          = 0;
        info->rep_ts          = 0;
        info->round           = 0;
        info->mode            = M_INITRTT;
        info->force_n_mode    = 0;
        return 0;
}


void send_packet(socket_udp *socket, struct shared *info, struct pkt_dtg *packet) {
        int ret;
        packet->mseq    = htonl(info->packets_sent);
        packet->dseq    = 0; /* unused */
        packet->ts      = htonl(get_time() / MS_TO_US);
        packet->rtt     = htonl(info->rtt_cur);
        packet->bitrate = htonl((u_int32) info->b_act);
        packet->round   = htonl(info->round);
        packet->mode    = htonl(info->mode);
        if ((ret = udp_send(socket, (char *) packet, PKT_SIZE)) < 0) {
                /* treat as lost packet, don't retransmit */
                fprintf(stderr, "%5.4f ", (float) get_time() / MICROSEC);
                perror("udp_send");
                if (info->mode == M_SLOWSTART) { /* kick out of SLOWSTART, we probably filled up the send buffer */
                        /* XXX next receiver report will get us back to SLOWSTART!!! fix that! */
                        info->mode = M_NORMAL;
                        info->b_exp /= 2;
                        info->b_act = info->b_exp;
                        assert(info->b_act != 0);

                        info->rate_delta = 0;
                        info->max_increase = info->b_act;
                        info->adj_ts = 0;
                        info->force_n_mode = 1; /* prevent going back to SLOWSTART when next report arrives */
                        printf("-> congestion control mode (send error)\n");
                }
        }
}


int send_loop(socket_udp *socket, struct shared *info, int duration, int cbr) {
        int buflen = PKT_SIZE;
        char buffer[buflen];
        struct pkt_dtg *packet = (struct pkt_dtg *) buffer;
        int64 start = 0, end = 0;
        int64 delay = 0, delay_diff = 0;
        int64 sim_end;
        int nr = 0;
        double modified_bact = 0;

        sim_end = get_time() + duration * MICROSEC;
        while ((nr < MAX_TOTAL_PACKETS) && (duration == 0 || get_time() < sim_end)) {
                start = get_time();
                ++nr;

                if (cbr == 0) {
                        switch (info->mode) {
                        case M_INITRTT:
                                break;
                        case M_SLOWSTART:
                                /* inc by delta over one rtt */
                                if (start > info->adj_ts + info->rtt * MS_TO_US) {
                                        if (info->b_rep > 0) {
                                                info->b_exp = min(2 * info->b_rep, 2 * info->b_act);
                                        }
                                        if (info->b_exp > info->max_bitrate)
                                                info->b_exp = info->max_bitrate;
                                        /* use of b_exp for interval is conservative */
                                        assert(info->b_exp != 0);
                                        assert(info->rtt != 0);
                                        info->rate_delta = (info->b_exp - info->b_act);

                                        /* and spread increase over packets sent during the next RTT */
                                        info->rate_delta /= max((info->b_act + info->rate_delta / 2) / PKT_SIZE * (info->rtt / MILLISEC), 1);

                                        if (info->rate_delta < 0.0) 
                                                info->rate_delta = 0.0;
                                        else
                                                info->adj_ts = start;
                                }
                                if (info->b_act + info->rate_delta < info->b_exp)
                                        info->b_act += info->rate_delta;
                                else
                                        info->b_act = info->b_exp;
                                break;
                                assert(info->b_act != 0);

                        case M_NORMAL:
                                /* inc smoothly each time a packet is sent */
                                info->b_act = min(info->b_act + info->rate_delta, info->max_increase);
                                assert(info->b_act != 0);
                                break;
                        default:
                                fprintf(stderr, "error: invalid mode (%i)\n", info->mode);
                                break;
                        }
                } else {
                        info->b_act = info->b_exp = cbr;
                }

                assert(info->b_act != 0);

                send_packet(socket, info, packet);
                ++info->packets_sent;

                /* ... and wait (to send at specified bitrate) */
                end = get_time();

                /* modify sending rate, should not influence Internet experiments but
                   improve Dummynet behavior */
                if (info->rtt_cur != 0 && info->mode == M_NORMAL) {
                        modified_bact = info->b_act * info->sqrt_rtt / sqrt(info->rtt_cur);
                        /* printf("%6.3f: %3.2f \t %3.2f \t %3.2f \t %3.2f\n", (double) start / MICROSEC, */
                               /* info->b_act, modified_bact, info->sqrt_rtt, sqrt(info->rtt_cur)); */
                } else {
                        modified_bact = info->b_act;
                }

                delay  = (int64) PKT_SIZE * MICROSEC / modified_bact; /* the time it should take */
                /* delay  = (int64) PKT_SIZE * MICROSEC / info->b_act * d_factor; */
                delay -= (end - start);                               /* minus the time it took */
                delay += delay_diff;                                  /* plus diff of last packet */
                if (delay > sim_end - end)
                        break;
                wait(delay);

                delay_diff = delay - (get_time() - end); /* waited too long / too short */
        }

        if (nr == MAX_TOTAL_PACKETS)
                fprintf(stderr, "maximum number of packets (%i) reached\n", MAX_TOTAL_PACKETS);
        return 0;
}


void update_mode(struct shared *info, u_int32 b_exp) {
        switch (info->mode) {
        case M_INITRTT:
                if (info->rtt > 0.0) {
                        info->mode = M_SLOWSTART;
                        info->b_act = PKT_SIZE / (info->rtt / MILLISEC) / 2;
                        info->b_rep = PKT_SIZE / (info->rtt / MILLISEC) / 2;
                        printf("%f slowstart mode\n", (float) info->rep_ts / MILLISEC);
                }
                break;
        case M_SLOWSTART:
                if (b_exp > 0) {
                        info->mode = M_NORMAL;
                        info->adj_ts = 0;
                        printf("%f congestion control mode\n", (float) info->rep_ts / MILLISEC);
                }
                break;
        case M_NORMAL:
                /* back to SLOWSTART mode when b_exp == 0 (that's when l_exp == 0) */
                if (b_exp == 0 && info->force_n_mode == 0) {
                        info->b_exp = 0;
                        info->mode = M_SLOWSTART;
                        info->adj_ts = 0;
                        printf("%f reentered slowstart mode\n", (float) info->rep_ts / MILLISEC);
                }
                break;
        default:
                fprintf(stderr, "error: invalid mode (%i)\n", info->mode);
                exit(1);
        }
}


void update_bexp(struct shared *info, u_int32 b_exp) {
        if (info->mode == M_NORMAL && b_exp > 0) {
                info->b_exp = (double) b_exp;
                info->b_exp = max(info->b_exp, info->min_bitrate);
                info->b_exp = min(info->b_exp, info->max_bitrate);
                
                if (info->b_act >= info->b_exp) {
                        info->b_act = info->b_exp;
                        info->max_increase = info->b_exp;
                        info->rate_delta = 0;
                        assert(info->max_increase > 0);
                } else {
                        /* inc by 1 packet */
                        info->rate_delta  = PKT_SIZE / (info->rtt / MILLISEC);
                        /* limit total increase to 2 packets/rtt until new rec. reports arrive */
                        info->max_increase = min(info->b_exp, info->b_act + 2 * info->rate_delta);
                        /* and spread increase over the approx. number of packets sent during the next RTT */
                        info->rate_delta /= (info->b_act + info->rate_delta / 2) / PKT_SIZE * (info->rtt / MILLISEC);
                        assert(info->max_increase > 0);
                }

                /* printf("%f: %f + %f -> %f\n", (float) info->rep_ts / MILLISEC, */
                /* info->b_act, info->rate_delta, info->b_exp);  */
        }
}


int recv_reports(socket_udp *socket, struct shared *info) {
        char buffer[REP_SIZE];
        struct rep_dtg *rep = (struct rep_dtg *) buffer;
        struct timeval max_wait_tv;
        double max_wait;
        int nready;

        udp_fd_zero();
        while (1) {
                udp_fd_set(socket);
                switch (info->mode) {
                case M_INITRTT:
                        /* assume 500ms RTT for timeout while in M_INITRTT */
                        max_wait = DEC_NO_REPORT * 500.0 / MILLISEC / REPORTS_PER_RTT;
                        break;
                case M_SLOWSTART:
                        /* use SLOWSTART_NO_REPORT instead of DEC_NO_REPORT, be more cautious */
                        max_wait = (double) SLOWSTART_NO_REPORT * 
                                max(max(info->rtt_cur, info->rtt) / MILLISEC / REPORTS_PER_RTT,
                                    PKT_SIZE / info->b_act);
                        break;
                case M_NORMAL:
                        /* use max (RTT_CUR, RTT, time to send one packet)
                           -> avoid decrease caused by heavy RTT variations */
                        max_wait = (double) DEC_NO_REPORT * 
                                max(max(info->rtt_cur, info->rtt) / MILLISEC / REPORTS_PER_RTT,
                                    PKT_SIZE / info->b_act);
                        break;
                default:
                        fprintf(stderr, "error: invalid mode (%i)\n", info->mode);
                        exit(1);
                }
                max_wait_tv.tv_sec = (long) max_wait;
                max_wait_tv.tv_usec = ((long) (max_wait * MICROSEC)) % MICROSEC;
                if ((nready = udp_select(&max_wait_tv)) <= 0) {
                        if (nready == 0) {
                                /* no receiver report for DEC_NO_REPORT x "expected time interval" 
                                 * -> assume congestion and reduce rate
                                 */
                                printf("%f timeout: no report for %3.2fms\n", (float) info->rep_ts / MILLISEC,
                                       max_wait * MILLISEC);
                                
                                if (info->mode == M_SLOWSTART) {
                                        info->mode = M_NORMAL;
                                        /* set adj_ts to 0 so that rate_delta is reset to an appropriate value
                                           as soon as we get out of slowstart */
                                        info->adj_ts = 0;
                                        /* keep sending at the current rate */
                                        info->b_exp = info->b_act;
                                        info->rate_delta = 0;
                                        info->max_increase = info->b_act;
                                        printf("%f congestion control mode\n", (float) info->rep_ts / MILLISEC);
                                } else if (info->mode == M_NORMAL) {
                                        info->b_act = max(info->b_act / 2, info->min_bitrate);
                                        info->max_increase = info->b_act;
                                        info->rate_delta = 0;
                                        info->b_exp = info->b_act;
                                }

                                assert(info->max_increase > 0);
                                assert(info->b_act > 0);

                                ++info->round; /* sender reacted by halving bitrate -> start new round*/
                                
                                continue;
                        } else if (errno == EINTR) {
                                continue;
                        } else {
                                perror("select_error");
                        }
                }
                if (udp_fd_isset(socket)) {
                        if (udp_recv(socket, buffer, REP_SIZE)) {
                                info->rep_ts = get_time() / MS_TO_US;
                                info->rtt_cur = info->rep_ts - ntohl(rep->ts);
                                ++info->round; /* new report -> start new round for loss measurement */
                                if (info->rtt > 0.0)
                                        info->rtt = (1 - RTT_DECAY) * info->rtt + RTT_DECAY * info->rtt_cur;
                                else
                                        info->rtt = info->rtt_cur;
                                
                                if (info->sqrt_rtt > 0.0)
                                        info->sqrt_rtt = (1 - RTT_DECAY) * info->sqrt_rtt + RTT_DECAY * sqrt(info->rtt_cur);
                                else
                                        info->sqrt_rtt = sqrt(info->rtt_cur);
                                
                                ++info->reports_recv;
                                info->b_rep = ntohl(rep->b_rep);
                                update_mode(info, ntohl(rep->b_exp));
                                update_bexp(info, ntohl(rep->b_exp));
                        }
                }
        }
        return 0;
}


int main(int argc, char *argv[]) {
        char *addr = NULL;
        int port = -1;
        int ttl = 16;
	int pid_recv;
        socket_udp *send_socket;
        socket_udp *recv_socket;
	char   *shmptr;
	int    shmid;
	struct shmid_ds buf;
        struct shared *info;
        int    duration = 0;
        int    cbr = 0;
        double max_bitrate = MAX_BITRATE;
        double min_bitrate = 0;
        /* double eqn_inc = EQN_INC; */
        /* double eqn_k = EQN_K; */
        /* double eqn_e = EQN_E; */
        int i;

        if (argc < 3) {
                fprintf(stderr, "usage: %s address port [ttl] [-t duration] [-m max. bitrate] [-i min. bitrate] [-c const. bitrate]\n", argv[0]);
                exit(1);
        } else {
                addr = argv[1];
                port = atoi(argv[2]);
                for (i = 3; i < argc; ++i) {
                        if (argv[i][0] != '-') {
                                ttl = atoi(argv[i]);
                        } else {
                                switch (argv[i][1]) {
                                case 't' : if (argc > i + 1) duration = atoi(argv[++i]); break;
                                case 'm' : if (argc > i + 1) max_bitrate = atof(argv[++i]); break;
                                case 'i' : if (argc > i + 1) min_bitrate = atof(argv[++i]); break;
                                        /* case 'p' : if (argc > i + 1) eqn_inc = atof(argv[++i]); break; */
                                        /* case 'k' : if (argc > i + 1) eqn_k = atof(argv[++i]); break; */
                                        /* case 'e' : if (argc > i + 1) eqn_e = atof(argv[++i]); break; */
                                case 'c' : if (argc > i + 1) cbr = atoi(argv[++i]); break;
                                default : fprintf(stderr, "unknown option: %s\n", argv[i]); exit(1); break;
                                }
                        }
                }
        }

        printf("started ");
        for (i = 0; i < argc; ++i)
                printf("%s ", argv[i]);
        printf("\n");

        send_socket = udp_init(addr, port  , ttl);
        recv_socket = udp_init(addr, port+1, ttl);
        assert(send_socket != NULL);
        assert(recv_socket != NULL);

        shmid = shmget (IPC_PRIVATE, sizeof(struct shared), SHM_R | SHM_W) ;
	if (shmid == -1) {
		perror("shmget");
		exit(1);
	}
        shmptr = shmat(shmid,0,0);
	if (shmptr == NULL) {
		perror("shmat");
		exit(1);
	}

        info = (struct shared *) shmptr;

        shm_init(info, min_bitrate, max_bitrate);
	if (shmid == -1) {
		perror("shm_init");
		exit(1);
	}
        shmctl(shmid, IPC_RMID, &buf);

        init_time();

	pid_recv = fork ();
        if (pid_recv == 0) {
                recv_reports(recv_socket, info);
        } else {
                send_loop(send_socket, info, duration, cbr);
                kill(pid_recv, SIGKILL);
        }
        shmdt(shmptr);
        printf("sender finished\n");

        exit(0);
}
